"""Distributions module."""
from UQpy.Distributions.collection.beta import Beta
from UQpy.Distributions.collection.binomial import Binomial
from UQpy.Distributions.collection.cauchy import Cauchy
from UQpy.Distributions.collection.chi_square import ChiSquare
from UQpy.Distributions.collection.exponential import Exponential
from UQpy.Distributions.collection.gamma import Gamma
from UQpy.Distributions.collection.gen_extreme import GenExtreme
from UQpy.Distributions.collection.inverse_gauss import InvGauss
from UQpy.Distributions.collection.laplace import Laplace
from UQpy.Distributions.collection.levy import Levy
from UQpy.Distributions.collection.logistic import Logistic
from UQpy.Distributions.collection.lognormal import Lognormal
from UQpy.Distributions.collection.maxwell import Maxwell
from UQpy.Distributions.collection.multinomial import Multinomial
from UQpy.Distributions.collection.mvnormal import MVNormal
from UQpy.Distributions.collection.normal import Normal
from UQpy.Distributions.collection.pareto import Pareto
from UQpy.Distributions.collection.poisson import Poisson
from UQpy.Distributions.collection.rayleigh import Rayleigh
from UQpy.Distributions.collection.truncated_normal import TruncNorm
from UQpy.Distributions.collection.uniform import Uniform
from UQpy.Distributions.collection.joint_ind import JointInd
from UQpy.Distributions.collection.joint_copula import JointCopula
